@@ -1,7 +1,6 @@
 /* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+/* clients/kinit/kinit.c - Initialize a credential cache */
 /*
- * clients/kinit/kinit.c
- *
  * Copyright 1990, 2008 by the Massachusetts Institute of Technology.
  * All Rights Reserved.
  *
@@ -23,15 +22,14 @@
  * M.I.T. makes no representations about the suitability of
  * this software for any purpose.  It is provided "as is" without express
  * or implied warranty.
- *
- *
- * Initialize a credentials cache.
  */
 
 #include "autoconf.h"
+#include <k5-int.h>
 #include "k5-platform.h"        /* for asprintf */
 #include <krb5.h>
 #include "extern.h"
+#include <locale.h>
 #include <string.h>
 #include <stdio.h>
 #include <time.h>
@@ -119,10 +117,12 @@ struct k_opts
     char* principal_name;
     char* service_name;
     char* keytab_name;
-    char* k5_cache_name;
+    char* k5_in_cache_name;
+    char* k5_out_cache_name;
     char *armor_ccache;
 
     action_type action;
+    int use_client_keytab;
 
     int num_pa_opts;
     krb5_gic_opt_pa_data *pa_opts;
@@ -134,9 +134,10 @@ struct k_opts
 struct k5_data
 {
     krb5_context ctx;
-    krb5_ccache cc;
+    krb5_ccache in_cc, out_cc;
     krb5_principal me;
     char* name;
+    krb5_boolean switch_to_cache;
 };
 
 #ifdef GETOPT_LONG
@@ -199,7 +200,7 @@ usage()
             "[-E" USAGE_LONG_ENTERPRISE "] "
             USAGE_BREAK
             "[-v] [-R] "
-            "[-k [-t keytab_file]] "
+            "[-k [-i|-t keytab_file]] "
             "[-c cachename] "
             USAGE_BREAK
             "[-S service_name] [-T ticket_armor_cache]"
@@ -209,27 +210,28 @@ usage()
             progname);
 
     fprintf(stderr, "    options:");
-    fprintf(stderr, "\t-V verbose\n");
-    fprintf(stderr, "\t-l lifetime\n");
-    fprintf(stderr, "\t-s start time\n");
-    fprintf(stderr, "\t-r renewable lifetime\n");
-    fprintf(stderr, "\t-f forwardable\n");
-    fprintf(stderr, "\t-F not forwardable\n");
-    fprintf(stderr, "\t-p proxiable\n");
-    fprintf(stderr, "\t-P not proxiable\n");
-    fprintf(stderr, "\t-n anonymous\n");
-    fprintf(stderr, "\t-a include addresses\n");
-    fprintf(stderr, "\t-A do not include addresses\n");
-    fprintf(stderr, "\t-v validate\n");
-    fprintf(stderr, "\t-R renew\n");
-    fprintf(stderr, "\t-C canonicalize\n");
-    fprintf(stderr, "\t-E client is enterprise principal name\n");
-    fprintf(stderr, "\t-k use keytab\n");
-    fprintf(stderr, "\t-t filename of keytab to use\n");
-    fprintf(stderr, "\t-c Kerberos 5 cache name\n");
-    fprintf(stderr, "\t-S service\n");
-    fprintf(stderr, "\t-T armor credential cache\n");
-    fprintf(stderr, "\t-X <attribute>[=<value>]\n");
+    fprintf(stderr, _("\t-V verbose\n"));
+    fprintf(stderr, _("\t-l lifetime\n"));
+    fprintf(stderr, _("\t-s start time\n"));
+    fprintf(stderr, _("\t-r renewable lifetime\n"));
+    fprintf(stderr, _("\t-f forwardable\n"));
+    fprintf(stderr, _("\t-F not forwardable\n"));
+    fprintf(stderr, _("\t-p proxiable\n"));
+    fprintf(stderr, _("\t-P not proxiable\n"));
+    fprintf(stderr, _("\t-n anonymous\n"));
+    fprintf(stderr, _("\t-a include addresses\n"));
+    fprintf(stderr, _("\t-A do not include addresses\n"));
+    fprintf(stderr, _("\t-v validate\n"));
+    fprintf(stderr, _("\t-R renew\n"));
+    fprintf(stderr, _("\t-C canonicalize\n"));
+    fprintf(stderr, _("\t-E client is enterprise principal name\n"));
+    fprintf(stderr, _("\t-k use keytab\n"));
+    fprintf(stderr, _("\t-i use default client keytab (with -k)\n"));
+    fprintf(stderr, _("\t-t filename of keytab to use\n"));
+    fprintf(stderr, _("\t-c Kerberos 5 cache name\n"));
+    fprintf(stderr, _("\t-S service\n"));
+    fprintf(stderr, _("\t-T armor credential cache\n"));
+    fprintf(stderr, _("\t-X <attribute>[=<value>]\n"));
     exit(2);
 }
 
@@ -286,8 +288,8 @@ parse_options(argc, argv, opts)
     int errflg = 0;
     int i;
 
-    while ((i = GETOPT(argc, argv, "r:fpFPn54aAVl:s:c:kt:T:RS:vX:CE"))
-           != -1) {
+    while ((i = GETOPT(argc, argv,
+                       "r:fpFPn54aAVl:s:c:kit:T:RS:vX:CEI:")) != -1) {
         switch (i) {
         case 'V':
             opts->verbose = 1;
@@ -296,7 +298,7 @@ parse_options(argc, argv, opts)
             /* Lifetime */
             code = krb5_string_to_deltat(optarg, &opts->lifetime);
             if (code != 0 || opts->lifetime == 0) {
-                fprintf(stderr, "Bad lifetime value %s\n", optarg);
+                fprintf(stderr, _("Bad lifetime value %s\n"), optarg);
                 errflg++;
             }
             break;
@@ -304,7 +306,7 @@ parse_options(argc, argv, opts)
             /* Renewable Time */
             code = krb5_string_to_deltat(optarg, &opts->rlife);
             if (code != 0 || opts->rlife == 0) {
-                fprintf(stderr, "Bad lifetime value %s\n", optarg);
+                fprintf(stderr, _("Bad lifetime value %s\n"), optarg);
                 errflg++;
             }
             break;
@@ -332,11 +334,13 @@ parse_options(argc, argv, opts)
         case 's':
             code = krb5_string_to_deltat(optarg, &opts->starttime);
             if (code != 0 || opts->starttime == 0) {
+                /* Parse as an absolute time; intentionally undocumented
+                 * but left for backwards compatibility. */
                 krb5_timestamp abs_starttime;
 
                 code = krb5_string_to_timestamp(optarg, &abs_starttime);
                 if (code != 0 || abs_starttime == 0) {
-                    fprintf(stderr, "Bad start time value %s\n", optarg);
+                    fprintf(stderr, _("Bad start time value %s\n"), optarg);
                     errflg++;
                 } else {
                     opts->starttime = abs_starttime - time(0);
@@ -349,10 +353,13 @@ parse_options(argc, argv, opts)
         case 'k':
             opts->action = INIT_KT;
             break;
+        case 'i':
+            opts->use_client_keytab = 1;
+            break;
         case 't':
             if (opts->keytab_name)
             {
-                fprintf(stderr, "Only one -t option allowed.\n");
+                fprintf(stderr, _("Only one -t option allowed.\n"));
                 errflg++;
             } else {
                 opts->keytab_name = optarg;
@@ -360,7 +367,7 @@ parse_options(argc, argv, opts)
             break;
         case 'T':
             if (opts->armor_ccache) {
-                fprintf(stderr, "Only one armor_ccache\n");
+                fprintf(stderr, _("Only one armor_ccache\n"));
                 errflg++;
             } else opts->armor_ccache = optarg;
             break;
@@ -371,19 +378,27 @@ parse_options(argc, argv, opts)
             opts->action = VALIDATE;
             break;
         case 'c':
-            if (opts->k5_cache_name)
+            if (opts->k5_out_cache_name)
             {
-                fprintf(stderr, "Only one -c option allowed\n");
+                fprintf(stderr, _("Only one -c option allowed\n"));
                 errflg++;
             } else {
-                opts->k5_cache_name = optarg;
+                opts->k5_out_cache_name = optarg;
+            }
+            break;
+        case 'I':
+            if (opts->k5_in_cache_name) {
+                fprintf(stderr, _("Only one -I option allowed\n"));
+                errflg++;
+            } else {
+                opts->k5_in_cache_name = optarg;
             }
             break;
         case 'X':
             code = add_preauth_opt(opts, optarg);
             if (code)
             {
-                com_err(progname, code, "while adding preauth option");
+                com_err(progname, code, _("while adding preauth option"));
                 errflg++;
             }
             break;
@@ -394,7 +409,7 @@ parse_options(argc, argv, opts)
             opts->enterprise = 1;
             break;
         case '4':
-            fprintf(stderr, "Kerberos 4 is no longer supported\n");
+            fprintf(stderr, _("Kerberos 4 is no longer supported\n"));
             exit(3);
             break;
         case '5':
@@ -407,22 +422,33 @@ parse_options(argc, argv, opts)
 
     if (opts->forwardable && opts->not_forwardable)
     {
-        fprintf(stderr, "Only one of -f and -F allowed\n");
+        fprintf(stderr, _("Only one of -f and -F allowed\n"));
         errflg++;
     }
     if (opts->proxiable && opts->not_proxiable)
     {
-        fprintf(stderr, "Only one of -p and -P allowed\n");
+        fprintf(stderr, _("Only one of -p and -P allowed\n"));
         errflg++;
     }
     if (opts->addresses && opts->no_addresses)
     {
-        fprintf(stderr, "Only one of -a and -A allowed\n");
+        fprintf(stderr, _("Only one of -a and -A allowed\n"));
         errflg++;
     }
+    if (opts->keytab_name != NULL && opts->use_client_keytab == 1)
+    {
+        fprintf(stderr, _("Only one of -t and -i allowed\n"));
+        errflg++;
+    }
+    if ((opts->keytab_name != NULL || opts->use_client_keytab == 1) &&
+        opts->action != INIT_KT)
+    {
+        opts->action = INIT_KT;
+        fprintf(stderr, _("keytab specified, forcing -k\n"));
+    }
 
     if (argc - optind > 1) {
-        fprintf(stderr, "Extra arguments (starting with \"%s\").\n",
+        fprintf(stderr, _("Extra arguments (starting with \"%s\").\n"),
                 argv[optind+1]);
         errflg++;
     }
@@ -441,124 +467,208 @@ k5_begin(opts, k5)
     struct k5_data* k5;
 {
     krb5_error_code code = 0;
+    int success = 0;
     int flags = opts->enterprise ? KRB5_PRINCIPAL_PARSE_ENTERPRISE : 0;
+    krb5_ccache defcache = NULL;
+    krb5_principal defcache_princ = NULL, princ;
+    krb5_keytab keytab;
+    const char *deftype = NULL;
+    char *defrealm, *name;
 
     code = krb5_init_context(&k5->ctx);
     if (code) {
-        com_err(progname, code, "while initializing Kerberos 5 library");
+        com_err(progname, code, _("while initializing Kerberos 5 library"));
         return 0;
     }
     errctx = k5->ctx;
-    if (opts->k5_cache_name)
-    {
-        code = krb5_cc_resolve(k5->ctx, opts->k5_cache_name, &k5->cc);
+
+    if (opts->k5_out_cache_name) {
+        code = krb5_cc_resolve(k5->ctx, opts->k5_out_cache_name, &k5->out_cc);
         if (code != 0) {
-            com_err(progname, code, "resolving ccache %s",
-                    opts->k5_cache_name);
-            return 0;
+            com_err(progname, code, _("resolving ccache %s"),
+                    opts->k5_out_cache_name);
+            goto cleanup;
         }
         if (opts->verbose) {
-            fprintf(stderr, "Using specified cache: %s\n",
-                    opts->k5_cache_name);
+            fprintf(stderr, _("Using specified cache: %s\n"),
+                    opts->k5_out_cache_name);
         }
+    } else {
+        /* Resolve the default ccache and get its type and default principal
+         * (if it is initialized). */
+        code = krb5_cc_default(k5->ctx, &defcache);
+        if (code) {
+            com_err(progname, code, _("while getting default ccache"));
+            goto cleanup;
+        }
+        deftype = krb5_cc_get_type(k5->ctx, defcache);
+        if (krb5_cc_get_principal(k5->ctx, defcache, &defcache_princ) != 0)
+            defcache_princ = NULL;
     }
-    else
-    {
-        if ((code = krb5_cc_default(k5->ctx, &k5->cc))) {
-            com_err(progname, code, "while getting default ccache");
-            return 0;
+
+    /* Choose a client principal name. */
+    if (opts->principal_name != NULL) {
+        /* Use the specified principal name. */
+        code = krb5_parse_name_flags(k5->ctx, opts->principal_name, flags,
+                                     &k5->me);
+        if (code) {
+            com_err(progname, code, _("when parsing name %s"),
+                    opts->principal_name);
+            goto cleanup;
         }
-        if (opts->verbose) {
-            fprintf(stderr, "Using default cache: %s\n",
-                    krb5_cc_get_name(k5->ctx, k5->cc));
+    } else if (opts->anonymous) {
+        /* Use the anonymous principal for the local realm. */
+        code = krb5_get_default_realm(k5->ctx, &defrealm);
+        if (code) {
+            com_err(progname, code, _("while getting default realm"));
+            goto cleanup;
+        }
+        code = krb5_build_principal_ext(k5->ctx, &k5->me,
+                                        strlen(defrealm), defrealm,
+                                        strlen(KRB5_WELLKNOWN_NAMESTR),
+                                        KRB5_WELLKNOWN_NAMESTR,
+                                        strlen(KRB5_ANONYMOUS_PRINCSTR),
+                                        KRB5_ANONYMOUS_PRINCSTR,
+                                        0);
+        krb5_free_default_realm(k5->ctx, defrealm);
+        if (code) {
+            com_err(progname, code, _("while building principal"));
+            goto cleanup;
+        }
+    } else if (opts->action == INIT_KT && opts->use_client_keytab) {
+        /* Use the first entry from the client keytab. */
+        code = krb5_kt_client_default(k5->ctx, &keytab);
+        if (code) {
+            com_err(progname, code,
+                    _("When resolving the default client keytab"));
+            goto cleanup;
+        }
+        code = k5_kt_get_principal(k5->ctx, keytab, &k5->me);
+        krb5_kt_close(k5->ctx, keytab);
+        if (code) {
+            com_err(progname, code,
+                    _("When determining client principal name from keytab"));
+            goto cleanup;
+        }
+    } else if (opts->action == INIT_KT) {
+        /* Use the default host/service name. */
+        code = krb5_sname_to_principal(k5->ctx, NULL, NULL, KRB5_NT_SRV_HST,
+                                       &k5->me);
+        if (code) {
+            com_err(progname, code,
+                    _("when creating default server principal name"));
+            goto cleanup;
+        }
+        if (k5->me->realm.data[0] == 0) {
+            code = krb5_unparse_name(k5->ctx, k5->me, &k5->name);
+            if (code == 0) {
+                com_err(progname, KRB5_ERR_HOST_REALM_UNKNOWN,
+                        _("(principal %s)"), k5->name);
+            } else {
+                com_err(progname, KRB5_ERR_HOST_REALM_UNKNOWN,
+                        _("for local services"));
+            }
+            goto cleanup;
+        }
+    } else if (k5->out_cc != NULL) {
+        /* If the output ccache is initialized, use its principal. */
+        if (krb5_cc_get_principal(k5->ctx, k5->out_cc, &princ) == 0)
+            k5->me = princ;
+    } else if (defcache_princ != NULL) {
+        /* Use the default cache's principal, and use the default cache as the
+         * output cache. */
+        k5->out_cc = defcache;
+        defcache = NULL;
+        k5->me = defcache_princ;
+        defcache_princ = NULL;
+    }
+
+    /* If we still haven't chosen, use the local username. */
+    if (k5->me == NULL) {
+        name = get_name_from_os();
+        if (name == NULL) {
+            fprintf(stderr, _("Unable to identify user\n"));
+            goto cleanup;
+        }
+        code = krb5_parse_name_flags(k5->ctx, name, flags, &k5->me);
+        if (code) {
+            com_err(progname, code, _("when parsing name %s"),
+                    name);
+            goto cleanup;
         }
     }
 
-    if (opts->principal_name)
-    {
-        /* Use specified name */
-        if ((code = krb5_parse_name_flags(k5->ctx, opts->principal_name,
-                                          flags, &k5->me))) {
-            com_err(progname, code, "when parsing name %s",
+    if (k5->out_cc == NULL && krb5_cc_support_switch(k5->ctx, deftype)) {
+        /* Use an existing cache for the client principal if we can. */
+        code = krb5_cc_cache_match(k5->ctx, k5->me, &k5->out_cc);
+        if (code != 0 && code != KRB5_CC_NOTFOUND) {
+            com_err(progname, code, _("while searching for ccache for %s"),
                     opts->principal_name);
-            return 0;
+            goto cleanup;
+        }
+        if (code == 0) {
+            if (opts->verbose) {
+                fprintf(stderr, _("Using existing cache: %s\n"),
+                        krb5_cc_get_name(k5->ctx, k5->out_cc));
+            }
+            k5->switch_to_cache = 1;
+        } else if (defcache_princ != NULL) {
+            /* Create a new cache to avoid overwriting the initialized default
+             * cache. */
+            code = krb5_cc_new_unique(k5->ctx, deftype, NULL, &k5->out_cc);
+            if (code) {
+                com_err(progname, code, _("while generating new ccache"));
+                goto cleanup;
+            }
+            if (opts->verbose) {
+                fprintf(stderr, _("Using new cache: %s\n"),
+                        krb5_cc_get_name(k5->ctx, k5->out_cc));
+            }
+            k5->switch_to_cache = 1;
         }
     }
-    else
-    {
-        /* No principal name specified */
-        if (opts->anonymous) {
-            char *defrealm;
-            code = krb5_get_default_realm(k5->ctx, &defrealm);
-            if (code) {
-                com_err(progname, code, "while getting default realm");
-                return 0;
-            }
-            code = krb5_build_principal_ext(k5->ctx, &k5->me,
-                                            strlen(defrealm), defrealm,
-                                            strlen(KRB5_WELLKNOWN_NAMESTR),
-                                            KRB5_WELLKNOWN_NAMESTR,
-                                            strlen(KRB5_ANONYMOUS_PRINCSTR),
-                                            KRB5_ANONYMOUS_PRINCSTR,
-                                            0);
-            krb5_free_default_realm(k5->ctx, defrealm);
-            if (code) {
-                com_err(progname, code, "while building principal");
-                return 0;
-            }
-        } else {
-            if (opts->action == INIT_KT) {
-                /* Use the default host/service name */
-                code = krb5_sname_to_principal(k5->ctx, NULL, NULL,
-                                               KRB5_NT_SRV_HST, &k5->me);
-                if (code) {
-                    com_err(progname, code,
-                            "when creating default server principal name");
-                    return 0;
-                }
-                if (k5->me->realm.data[0] == 0) {
-                    code = krb5_unparse_name(k5->ctx, k5->me, &k5->name);
-                    if (code == 0) {
-                        com_err(progname, KRB5_ERR_HOST_REALM_UNKNOWN,
-                                "(principal %s)", k5->name);
-                    } else {
-                        com_err(progname, KRB5_ERR_HOST_REALM_UNKNOWN,
-                                "for local services");
-                    }
-                    return 0;
-                }
-            } else {
-                /* Get default principal from cache if one exists */
-                code = krb5_cc_get_principal(k5->ctx, k5->cc,
-                                             &k5->me);
-                if (code) {
-                    char *name = get_name_from_os();
-                    if (!name) {
-                        fprintf(stderr, "Unable to identify user\n");
-                        return 0;
-                    }
-                    if ((code = krb5_parse_name_flags(k5->ctx, name,
-                                                      flags, &k5->me))) {
-                        com_err(progname, code, "when parsing name %s",
-                                name);
-                        return 0;
-                    }
-                }
-            }
+
+    /* Use the default cache if we haven't picked one yet. */
+    if (k5->out_cc == NULL) {
+        k5->out_cc = defcache;
+        defcache = NULL;
+        if (opts->verbose) {
+            fprintf(stderr, _("Using default cache: %s\n"),
+                    krb5_cc_get_name(k5->ctx, k5->out_cc));
         }
     }
 
+    if (opts->k5_in_cache_name) {
+        code = krb5_cc_resolve(k5->ctx, opts->k5_in_cache_name, &k5->in_cc);
+        if (code != 0) {
+            com_err(progname, code, _("resolving ccache %s"),
+                    opts->k5_in_cache_name);
+            goto cleanup;
+        }
+        if (opts->verbose) {
+            fprintf(stderr, _("Using specified input cache: %s\n"),
+                    opts->k5_in_cache_name);
+        }
+    }
+
+
     code = krb5_unparse_name(k5->ctx, k5->me, &k5->name);
     if (code) {
-        com_err(progname, code, "when unparsing name");
-        return 0;
+        com_err(progname, code, _("when unparsing name"));
+        goto cleanup;
     }
     if (opts->verbose)
-        fprintf(stderr, "Using principal: %s\n", k5->name);
+        fprintf(stderr, _("Using principal: %s\n"), k5->name);
 
     opts->principal_name = k5->name;
 
-    return 1;
+    success = 1;
+
+cleanup:
+    if (defcache != NULL)
+        krb5_cc_close(k5->ctx, defcache);
+    krb5_free_principal(k5->ctx, defcache_princ);
+    return success;
 }
 
 static void
@@ -569,8 +679,10 @@ k5_end(k5)
         krb5_free_unparsed_name(k5->ctx, k5->name);
     if (k5->me)
         krb5_free_principal(k5->ctx, k5->me);
-    if (k5->cc)
-        krb5_cc_close(k5->ctx, k5->cc);
+    if (k5->in_cc)
+        krb5_cc_close(k5->ctx, k5->in_cc);
+    if (k5->out_cc)
+        krb5_cc_close(k5->ctx, k5->out_cc);
     if (k5->ctx)
         krb5_free_context(k5->ctx);
     errctx = NULL;
@@ -637,7 +749,7 @@ k5_kinit(opts, k5)
         krb5_address **addresses = NULL;
         code = krb5_os_localaddr(k5->ctx, &addresses);
         if (code != 0) {
-            com_err(progname, code, "getting local addresses");
+            com_err(progname, code, _("getting local addresses"));
             goto cleanup;
         }
         krb5_get_init_creds_opt_set_address_list(options, addresses);
@@ -651,12 +763,12 @@ k5_kinit(opts, k5)
     if ((opts->action == INIT_KT) && opts->keytab_name)
     {
 #ifndef _WIN32
-        if (strncmp(opts->keytab_name, "KDB:", 3) == 0) {
+        if (strncmp(opts->keytab_name, "KDB:", 4) == 0) {
             code = kinit_kdb_init(&k5->ctx,
                                   krb5_princ_realm(k5->ctx, k5->me)->data);
             if (code != 0) {
                 com_err(progname, code,
-                        "while setting up KDB keytab for realm %s",
+                        _("while setting up KDB keytab for realm %s"),
                         krb5_princ_realm(k5->ctx, k5->me)->data);
                 goto cleanup;
             }
@@ -665,12 +777,18 @@ k5_kinit(opts, k5)
 
         code = krb5_kt_resolve(k5->ctx, opts->keytab_name, &keytab);
         if (code != 0) {
-            com_err(progname, code, "resolving keytab %s",
+            com_err(progname, code, _("resolving keytab %s"),
                     opts->keytab_name);
             goto cleanup;
         }
         if (opts->verbose)
-            fprintf(stderr, "Using keytab: %s\n", opts->keytab_name);
+            fprintf(stderr, _("Using keytab: %s\n"), opts->keytab_name);
+    } else if (opts->action == INIT_KT && opts->use_client_keytab) {
+        code = krb5_kt_client_default(k5->ctx, &keytab);
+        if (code != 0) {
+            com_err(progname, code, _("resolving default client keytab"));
+            goto cleanup;
+        }
     }
 
     for (i = 0; i < opts->num_pa_opts; i++) {
@@ -678,16 +796,23 @@ k5_kinit(opts, k5)
                                               opts->pa_opts[i].attr,
                                               opts->pa_opts[i].value);
         if (code != 0) {
-            com_err(progname, code, "while setting '%s'='%s'",
+            com_err(progname, code, _("while setting '%s'='%s'"),
                     opts->pa_opts[i].attr, opts->pa_opts[i].value);
             goto cleanup;
         }
         if (opts->verbose) {
-            fprintf(stderr, "PA Option %s = %s\n", opts->pa_opts[i].attr,
+            fprintf(stderr, _("PA Option %s = %s\n"), opts->pa_opts[i].attr,
                     opts->pa_opts[i].value);
         }
     }
-    code = krb5_get_init_creds_opt_set_out_ccache(k5->ctx, options, k5->cc);
+    if (k5->in_cc) {
+        code = krb5_get_init_creds_opt_set_in_ccache(k5->ctx, options,
+                                                     k5->in_cc);
+        if (code)
+            goto cleanup;
+    }
+    code = krb5_get_init_creds_opt_set_out_ccache(k5->ctx, options,
+                                                  k5->out_cc);
     if (code)
         goto cleanup;
 
@@ -707,11 +832,11 @@ k5_kinit(opts, k5)
                                           options);
         break;
     case VALIDATE:
-        code = krb5_get_validated_creds(k5->ctx, &my_creds, k5->me, k5->cc,
+        code = krb5_get_validated_creds(k5->ctx, &my_creds, k5->me, k5->out_cc,
                                         opts->service_name);
         break;
     case RENEW:
-        code = krb5_get_renewed_creds(k5->ctx, &my_creds, k5->me, k5->cc,
+        code = krb5_get_renewed_creds(k5->ctx, &my_creds, k5->me, k5->out_cc,
                                       opts->service_name);
         break;
     }
@@ -721,45 +846,53 @@ k5_kinit(opts, k5)
         switch (opts->action) {
         case INIT_PW:
         case INIT_KT:
-            doing = "getting initial credentials";
+            doing = _("getting initial credentials");
             break;
         case VALIDATE:
-            doing = "validating credentials";
+            doing = _("validating credentials");
             break;
         case RENEW:
-            doing = "renewing credentials";
+            doing = _("renewing credentials");
             break;
         }
 
         if (code == KRB5KRB_AP_ERR_BAD_INTEGRITY)
-            fprintf(stderr, "%s: Password incorrect while %s\n", progname,
+            fprintf(stderr, _("%s: Password incorrect while %s\n"), progname,
                     doing);
         else
-            com_err(progname, code, "while %s", doing);
+            com_err(progname, code, _("while %s"), doing);
         goto cleanup;
     }
 
     if ((opts->action != INIT_PW) && (opts->action != INIT_KT)) {
-        code = krb5_cc_initialize(k5->ctx, k5->cc, opts->canonicalize ?
+        code = krb5_cc_initialize(k5->ctx, k5->out_cc, opts->canonicalize ?
                                   my_creds.client : k5->me);
         if (code) {
-            com_err(progname, code, "when initializing cache %s",
-                    opts->k5_cache_name?opts->k5_cache_name:"");
+            com_err(progname, code, _("when initializing cache %s"),
+                    opts->k5_out_cache_name?opts->k5_out_cache_name:"");
             goto cleanup;
         }
         if (opts->verbose)
-            fprintf(stderr, "Initialized cache\n");
+            fprintf(stderr, _("Initialized cache\n"));
 
-        code = krb5_cc_store_cred(k5->ctx, k5->cc, &my_creds);
+        code = krb5_cc_store_cred(k5->ctx, k5->out_cc, &my_creds);
         if (code) {
-            com_err(progname, code, "while storing credentials");
+            com_err(progname, code, _("while storing credentials"));
             goto cleanup;
         }
         if (opts->verbose)
-            fprintf(stderr, "Stored credentials\n");
+            fprintf(stderr, _("Stored credentials\n"));
     }
     notix = 0;
 
+    if (k5->switch_to_cache) {
+        code = krb5_cc_switch(k5->ctx, k5->out_cc);
+        if (code) {
+            com_err(progname, code, _("while switching to new ccache"));
+            goto cleanup;
+        }
+    }
+
 cleanup:
     if (options)
         krb5_get_init_creds_opt_free(k5->ctx, options);
@@ -786,6 +919,7 @@ main(argc, argv)
     struct k5_data k5;
     int authed_k5 = 0;
 
+    setlocale(LC_ALL, "");
     progname = GET_PROGNAME(argv[0]);
 
     /* Ensure we can be driven from a pipe */
@@ -809,7 +943,7 @@ main(argc, argv)
         authed_k5 = k5_kinit(&opts, &k5);
 
     if (authed_k5 && opts.verbose)
-        fprintf(stderr, "Authenticated to Kerberos v5\n");
+        fprintf(stderr, _("Authenticated to Kerberos v5\n"));
 
     k5_end(&k5);
 
